// PatchWorks 1.5 AdminDB File
// This Software is released under the terms of the GPL.

// DataBase Administration



#include "_includes.h"
#include "_defines.h"
#include "_structs.h"

#include "Image.h"
#include "FileSysHelp.h"

int AddDBItem(char *, sDataBase *);
int CreateIndexFile(char *, uint, uint, uint, uint);
int ReadIndexFile(char *, sDataBase *);



int OpenDB(sDataBase DataBase[], char *DBpath)
{
	uint i, sof, dof; //  source offset, destination offset
	int  err, newDB=-1;
	char *ipath; // path of the DB-Index file

	for(i=0; i<MAX_OPEN_DBs; i++)
	{
		if(DataBase[i].inuse == 0)
		{
			newDB = i;
			break;
		}
	}
	if(newDB == -1) return 1;					// To many DBs open

	// now open the DB
	DataBase[newDB].path = malloc(strlen(DBpath) + 5);
	ipath                = malloc(strlen(DBpath) + 5 + strlen(DB_INDEX_FILE));
	if( DataBase[newDB].path == NULL ) return 2;
	if( ipath                == NULL ) return 2; // not enough memory?

	// obsolete. done by bslash2slash // strcpy(DataBase[newDB].path, DBpath);	// set DB.path

	BSlash2Slash(DataBase[newDB].path, DBpath);	// replace '\' with '/'
	strcpy(ipath, DataBase[newDB].path);

	if( IsDir(DataBase[newDB].path) ) return 3; // could not access directory

	PathMerger(ipath, DB_INDEX_FILE); // get path of DB index

	err=ReadIndexFile(ipath, &DataBase[newDB]);

	free(ipath);

	if(err) return 10 + err;			// error ( > 10 ) while reading the index file

	// get memory to count usage of patches in CreatePatchwork. calloc initializes memory with zeros.
	DataBase[newDB].usecount = calloc( DataBase[newDB].imgcount, sizeof(uint) );
	if( DataBase[newDB].usecount == NULL ) return 2;

	// get memory to store the average color data as integer
	// for faster access than on the ascii values in DataBase.data
	// getting imagecaount * 3 integer "cells"
	DataBase[newDB].colordata = (uint *) calloc( DataBase[newDB].imgcount*3, sizeof(uint) );
	if( DataBase[newDB].colordata == NULL ) return 2;

	// fill colordata
	for(i=0; i<DataBase[newDB].imgcount; i++)
	{
		sof = i*INDEX_FILE_LINE + 13;		// offset. 13 = first byte of RRR, GGG, BBB
		dof = i*3;							// source offset
		// sscanf is slower than atoi!
		//err = sscanf(&DataBase[newDB].data[o], "%i %i %i", &(	DataBase[newDB].colordata[so  ]), &(DataBase[newDB].colordata[so+1]), &(DataBase[newDB].colordata[so+2]) );
		//if(err != 3) return 8;				// could not read all channels

		DataBase[newDB].colordata[dof  ] = atoi(&DataBase[newDB].data[sof  ]);
		DataBase[newDB].colordata[dof+1] = atoi(&DataBase[newDB].data[sof+4]);
		DataBase[newDB].colordata[dof+2] = atoi(&DataBase[newDB].data[sof+8]);
	}
	// DB.path:		is set
	// DB.data:		is set (ReadIndexFile)
	// DB.imgcount	is set (ReadIndexFile)
	// DB.res_h		is set (ReadIndexFile)
	// DB.res_w		is set (ReadIndexFile)
	// DB.dpi		is set (ReadIndexFile)
	// DB.used		is set
	DataBase[newDB].changed = 0;
	DataBase[newDB].inuse   = 1;

	return 0;
}




int CloseDB(sDataBase DataBase[], uint oldDB)
{
	if( DataBase[oldDB].inuse == 0  ) return 1; // already closed / not in use
	if( DataBase[oldDB].changed == 1) return 2; // unsaved changes! save, or set changed=0 to override

	DataBase[oldDB].changed = 0;
	DataBase[oldDB].imgcount = 0;
	DataBase[oldDB].path = 0;
	free(DataBase[oldDB].data);
	free(DataBase[oldDB].path);
	free(DataBase[oldDB].usecount);
	free(DataBase[oldDB].colordata);
	DataBase[oldDB].inuse = 0;
	
	return 0;
}

int SaveDB(sDataBase DataBase[], uint sDB)
{
	FILE *fp;
	char *ipath; // path of the DB-Index file
	int err;

	if( ! DataBase[sDB].inuse   ) return 1; // DB not in use
	if( ! DataBase[sDB].changed ) return 2; // no changes to save

	ipath = malloc(strlen(DataBase[sDB].path) + 5 + strlen(DB_INDEX_FILE));
	if( ipath == NULL ) return 3;		// not enough memory?
	strcpy(ipath, DataBase[sDB].path);
	if( IsDir(ipath) ) return 4;		// directory does not exist?

	PathMerger(ipath, DB_INDEX_FILE);	// get path of DB index

	// Delete current index
	if (DelFile(ipath) ) return 5;		// could not delete file

	// Create new index file
	err = CreateIndexFile(ipath, DataBase[sDB].imgcount, DataBase[sDB].res_h, DataBase[sDB].res_w, DataBase[sDB].dpi);
	if(err) return 10 + err;			// return error code ( > 10 ) of CreateIndexFile()

	// Write Data
	fp = fopen(ipath, "a+");
	if( fp == NULL) return 6;			// could not create file
	else
	{
		fprintf(fp, "%s", DataBase[sDB].data);
	}

	free(ipath);

	DataBase[sDB].changed = 0;
	
	return 0;
}


// AddDB Syntax:
// DBpath:	DBpath/name
// res:	resolution in pixels (x/y)
// Returns:	0 (default)
// 			1 illegal resolution or dpi
// 			2 out of memory
// 			3 directory already exists
//			4 couldn't create directory
int AddDB(char *DBpath, uint res_x, uint res_y, uint dpi)
{
	char *path;
	int err;

	if( (res_x<PATCH_RES_MINIMUM) || (res_x>PATCH_RES_MAXIMUM) ) return 1; // illegal resolution
	if( (res_y<PATCH_RES_MINIMUM) || (res_y>PATCH_RES_MAXIMUM) ) return 1;
	if( (dpi  <PATCH_DPI_MINIMUM) || (dpi  >PATCH_DPI_MAXIMUM) ) return 1;

	path = malloc(strlen(DBpath) + strlen(DB_INDEX_FILE) + 5);
	if(path == NULL) return 2;			// not enough meomory

	BSlash2Slash(path, DBpath);

	if( ! IsDir(path) ) return 3;		// already exists
	if(   MkDir(path) ) return 4;		// couldn't create

	PathMerger(path, DB_INDEX_FILE); // get path of DB index

	err = CreateIndexFile(path, 0, res_x, res_y, dpi);
	if(err) return 10 + err; // return error code ( > 10 ) of CreateIndexFile()

	free(path);

	return 0;
}




int DeleteDB(sDataBase DataBase[], uint oldDB)
{
	int err;

	if( DataBase[oldDB].inuse == 0  ) return 1; // already closed / not in use
	if( DataBase[oldDB].changed == 1) return 2; // unsaved changes! save, or set changed=0 to override
	//same checks as in CloseDB()

	if( DeleteFolderContents(DataBase[oldDB].path) ) return 3;	// could not delete (all) items in folder
	if( RmDir(DataBase[oldDB].path)                ) return 4;	// could not remove folder

	err = CloseDB(DataBase, oldDB);

	if(err) return 10 + err;		// error ( > 10 ) while closing DB

	return 0;
}




int AddDBItems(char *path, sDataBase *DataBase, int print)
{
	char *lowcasepath = NULL, *tmplist, *filelist = NULL;
	uint err, notall=0, offset=0;

	//get lowercase name
	GetLowercase(&lowcasepath, path);

	if( strstr(lowcasepath, ".bmp") ) // is a single file
	{
		return AddDBItem(path, DataBase);
	}
	else						// is probably a directory
	{
		GetFileList(path, &filelist);
		tmplist = filelist;

		printf("starting batch conversion:\n");		
		
		while(1)
		{
			tmplist += offset;

			if( strstr(tmplist, FILELIST_END) )	break;

			if(print) printf(err5_AddDBItem, tmplist); // adding img...
			err = AddDBItem(tmplist, DataBase);
			if( err )
			{
				if(print) printf(err6_AddDBItem);		// error, not added
				if( err == -1 ) // image has not been added.
				{
					notall = 9; // not all images have been added
				}
				else return err;
			}

			offset = strlen(tmplist) + 1;
		}

		free(filelist);
	}

	return notall; // this is 0 or 9
}




int AddDBItem(char *imagefile, sDataBase *DataBase) 
{
	int check, err;
	char *patchfile, patchname[20];
	sImage image, patch;
	
	patchfile = malloc(strlen(DataBase->path) + DB_FILE_LEN + 5);
	if( patchfile == NULL ) return 1;		// not enough memory

	if( DataBase->inuse == 0 ) return 2;	 // illegal database specified
	if(DataBase->imgcount >= 99999 ) return 3; // DB is full

	sprintf(patchname, "img%05i.bmp", DataBase->imgcount + 1 );
	strcpy(patchfile, DataBase->path);
	PathMerger(patchfile, patchname);		// create full path

	err = ReadImage(&image,  imagefile);	if(err) return 10 + err; // errors in XXXImage()

	patch.w = DataBase->res_w;
	patch.h = DataBase->res_h;

	err = ScaleImage(&image, &patch);
	if(err)
	{
		if( err == 2 ) return -1; // file has not been added, because it is too small
		return 20 + err;
	}
	err = GetAverageColor(&patch);			if(err) return 30 + err;
	err = WriteImage(&patch, patchfile);	if(err) return 40 + err;


	// Memory for at least imgcount data sets
	DataBase->data = (char*)realloc( DataBase->data, 1024 + (DataBase->imgcount+1)*INDEX_FILE_LINE );
	if( DataBase->data == NULL ) return 1; // not enough memory

	check = sprintf(DataBase->data + DataBase->imgcount*INDEX_FILE_LINE, "%s %3i %3i %3i\n", patchname, patch.avgc_r, patch.avgc_g, patch.avgc_b);
	if( check != INDEX_FILE_LINE) return 4; // internal error 

	DataBase->changed++;	// count the changes that are made
	DataBase->imgcount++;

	return 0;
}




int DeleteDBItem(sDataBase *DataBase, char *item)
{
// to be done	

	return 1;

	return 0;
}






// SUB FUNCTIONS
// =============


// IndexFile structure:
// WARNING			\n
// DB_INDEX_VERSION (\n)
// IMG_COUNT		\n
// PATCH_RES_X		\n
// PATCH_RES_Y		\n
// DPI				\n
// pat00001.bmp RRR GGG BBB  [12 + 3*4 = 24 chars + \n] => INDEX_FILE_LINE = 25
// pat00002.bmp RRR GGG BBB 
// [...]





int CreateIndexFile(char *path, uint imgcount, uint res_x, uint res_y, uint dpi)
{
	FILE *fp;

	fp = fopen(path, "r" );
	if( fp != NULL)
	{
		fclose(fp);
		return 1;					// file already exists
	}
	else
	{	
		fp = fopen(path, "w");
		if( fp == NULL) return 2;	// could not create file
		else
		{
			fprintf(fp, "%s\n", info_DONT_EDIT);
			fprintf(fp, "%s", DB_INDEX_VERSION);
			fprintf(fp, "%i\n", imgcount);
			fprintf(fp, "%i\n", res_x);
			fprintf(fp, "%i\n", res_y);
			fprintf(fp, "%i\n", dpi);
			fclose(fp);
		}
	}

	return 0;
}


int ReadIndexFile(char *ipath, sDataBase *DataBase)
{
	uint offset=0;
	FILE *fp;

	fp = fopen(ipath, "r" );
	if( fp == NULL)	return 1;	// could not open file		

	// is a valid DB then (hopefully)

	DataBase->data = malloc(1024);
	if( DataBase->data == NULL ) return 2; // not enough memory


	fgets(DataBase->data, 1000, fp); // read WARNING line, which should not exceed 1000 chars
	fgets(DataBase->data, 1000, fp); // read DB_INDEX_VERSION line, which should not exceed 1000 chars

	if( strcmp(DataBase->data, DB_INDEX_VERSION) ) return 9; // Probably wrong DataBase Version

	fgets(DataBase->data, 1000, fp); // read IMG_COUNT line

	if( (sscanf(DataBase->data, "%i", &DataBase->imgcount) !=1) ) return 3; // IMG_COUNT not read, file may be corrupt
	if( (DataBase->imgcount < 0) || (DataBase->imgcount > MAX_IMAGES) ) return 3;

	fgets(DataBase->data, 1000, fp); // read PATCH_RES_X line
	if( (sscanf(DataBase->data, "%i", &DataBase->res_w) !=1) ) return 4; // illegal PATCH_RES_X
	if( (DataBase->res_w < PATCH_RES_MINIMUM) || (DataBase->res_w > PATCH_RES_MAXIMUM) ) return 4;

	fgets(DataBase->data, 1000, fp); // read PATCH_RES_Y line
	if( (sscanf(DataBase->data, "%i", &DataBase->res_h) !=1) ) return 5; // illegal PATCH_RES_Y
	if( (DataBase->res_h < PATCH_RES_MINIMUM) || (DataBase->res_h > PATCH_RES_MAXIMUM) ) return 5;

	fgets(DataBase->data, 1000, fp); // read DPI line
	if( (sscanf(DataBase->data, "%i", &DataBase->dpi) !=1) ) return 6; // illegal DPI
	if( (DataBase->dpi < PATCH_DPI_MINIMUM) || (DataBase->dpi > PATCH_DPI_MAXIMUM) ) return 6;

	// Memory for at least imgcount data sets
	DataBase->data = (char*)realloc(DataBase->data, 1024 + DataBase->imgcount*INDEX_FILE_LINE);
	if( DataBase->data == NULL ) return 2; // not enough memory

	while(1)
	{
		if( (fgets(DataBase->data+offset, 1000, fp)==NULL) ) break; // read IMG_COUNT line (should be INDEX_FILE_LINE long)
		if( strlen(DataBase->data+offset) !=                    INDEX_FILE_LINE ) return 7;	// file corrupted
		if( strlen(DataBase->data       ) >  DataBase->imgcount*INDEX_FILE_LINE ) return 8; // more entries than allowed
		//*(DataBase->data+offset-1) = '\0'; // change \n to end-of-string (not necessary... had to be reversed before save)
		offset += INDEX_FILE_LINE;	// the terminating \0 will be overwritten
	}

	fclose(fp);

	return 0;
}