// PatchWorks 1.5 FileSysHelp File
// This Software is released under the terms of the GPL.

// FileSystem Helper Functions



#include "_includes.h"
#include "_defines.h"

#include "FileSysHelp.h" // prototypes needed internally


#if defined(linux) || defined(__linux)
		//philippe
#elif defined(_WIN32) || defined(__WIN32__)
	#include <windows.h>
	
#else
	#error Unknown Operating System...
#endif




/*		REPLACED WITH BSlash2Slash()

// replaces the '\' with '\\' to avoid them beeing escape sequences.
// example: part1\part2 -> part1\ + \part2 = part1\\part 2
// doesn't matter in non Win-OS
// output path must be at least 2 times input lenght. (worst case)
int EscSlash(char *path, char *DBpath)
{
	int count = 1;
	char *tail = DBpath;
	int  offset;

	strcpy(path, DBpath);

	do
	{
		// printf("progress: %s\n", tail);
		tail = strstr(tail, "\x5C"); // compare for "\"
		if(tail)
		{
			offset = tail - DBpath + count++; // count: append one later each time
			strcpy(path + offset, tail);
			tail++; // don't check the just added "\"
		}
	}
	while(tail);

	return 0;
}
*/


// replaces the '\' with '/' to avoid them beeing escape sequences.
// necessary for Win: CreateDirectory()
int BSlash2Slash(char *path, char *DBpath)
{
	int  i, len;

	strcpy(path, DBpath);
	len = strlen(path);

	for(i=0; i<len; i++)
	{
		if(path[i] == '\\') path[i] = '/'; 
	}

	return 0;
}


// checks if the specified directory exists
// returns 0 on success, 1 on errors
int IsDir(char * path)
{
#if defined(linux) || defined(__linux)
		//philippe
		return 1;
#elif defined(_WIN32) || defined(__WIN32__)

	HANDLE fh;
	WIN32_FIND_DATA FindFileData;


	//printf("%s\n", path);

	fh = FindFirstFile(path, &FindFileData ); // the file returned is normally "." (not on root dir)

	if(fh == INVALID_HANDLE_VALUE )
	{
		//printf("Invalid File Handle. GetLastError reports %d\n", GetLastError() );
		return 1; // unexpected error
	}

#else
	#error Unknown Operating System...
#endif

	return 0;
}


// creates the specified directory OS dependant
// returns 0 on success, 1 on errors
int MkDir(char * path)
{
#if defined(linux) || defined(__linux)
		//philippe
		return 1;
#elif defined(_WIN32) || defined(__WIN32__)
	// Warning: beware of escape sequences
	//printf("%i code", CreateDirectory("C:\\newone", NULL));
	//printf("%i code", GetLastError());
	
	if(CreateDirectory(path, NULL) == 0) return 1; // couldn't create dir.

#else
	#error Unknown Operating System...
#endif

	return 0;
}

// removes the specified directory OS dependant
// returns 0 on success, 1 on errors
int RmDir(char *path)
{
#if defined(linux) || defined(__linux)
		//philippe
		return 1;
#elif defined(_WIN32) || defined(__WIN32__)
	// Warning: beware of escape sequences
	//printf("%i code", CreateDirectory("C:\\newone", NULL));
	//printf("%i code", GetLastError());
	
	if(RemoveDirectory(path) == 0) return 1; // couldn't remove dir.

#else
	#error Unknown Operating System...
#endif

	return 0;
}



// returns a list of files in a folder
int GetFileList(char *path, char **filelist)
{
	HANDLE fh;
	WIN32_FIND_DATA FindFileData;

	int i=0, success, err, filenamelength, listlength=0;
	char *filepath, *dirpath, *lowcasename = NULL;

	//DEBUG only
	//LPVOID lpMsgBuf;

	dirpath = malloc(strlen(path) + 20); // actually 2 more chars should be enough... ( "/*" )
	if(dirpath == NULL) return 1;

	filepath = malloc(strlen(path) + MAX_PATH + 5);
	if(filepath == NULL) return 1;

	//strcpy(dirpath, path);
	BSlash2Slash(dirpath, path);

	PathMerger(dirpath, "*");

	while(1)
	{
		if(i==0)
		{
			i++;
			fh = FindFirstFile(dirpath, &FindFileData ); // the file returned is normally "." (not on root dir)

			if(fh == INVALID_HANDLE_VALUE )
			{
				printf("Invalid File Handle. GetLastError reports %d\n", GetLastError() );
				return 1; // unexpected error
			}
		}
		else
		{
			success = FindNextFile(fh, &FindFileData);

			if( ! success )
			{
				
				if( GetLastError() == ERROR_NO_MORE_FILES ) break;
				else printf("Error. GetLastError reports %d\n", GetLastError() );
				return 2; // unexpected error
			}
		}

		//get lowercase name
		err = GetLowercase(&lowcasename, FindFileData.cFileName);
		if( err )
		{
			printf(eNO_MEMORY);
			return err;
		}

		// add to list if cFileName contains ".bmp" or ".BMP" or...
		if( strstr(lowcasename, ".bmp") )
		{
			//strcpy(filepath, path);
			BSlash2Slash(filepath, path);
			PathMerger(filepath, FindFileData.cFileName);

			filenamelength = 1 + strlen(filepath); // one more for '\0'

			*filelist = realloc(*filelist, listlength + filenamelength);
			if( *filelist == NULL ) return 3; //not enough memory
			
			strcpy(*filelist+listlength, filepath); // add to list

			listlength += filenamelength; // new listlength, index where next one will be added.
		}
	} // while

	// add end info
	*filelist = realloc(*filelist, listlength + strlen(FILELIST_END) + 5);
	if( *filelist == NULL ) return 3; //not enough memory

	strcpy(*filelist+listlength, FILELIST_END);

	if( FindClose(fh) == 0 ) return 4; // Close file search. / Unexpected Error

	free(lowcasename);
	free(filepath);
	free(dirpath);

	return 0;
}




// deletes all contents of a folder.
// returns 0 on success, 1,2,3 on errors
int DeleteFolderContents(char *path)
{
	HANDLE fh;
	WIN32_FIND_DATA FindFileData;

	int i=0, success;
	char *delpath, *dirpath;

	//DEBUG only
	LPVOID lpMsgBuf;

	dirpath = malloc(strlen(path) + 20); // actually 2 more chars should be enough... ( "/*" )
	if(dirpath == NULL) return 1;
	strcpy(dirpath, path);

	delpath = malloc(strlen(path) + MAX_PATH + 5);
	if(delpath == NULL) return 1;


	PathMerger(dirpath, "*");

	//DEBUG prints the path that is searched for
	//printf("%s no1\n", dirpath);

	while(1)
	{
		if(i==0)
		{
			i++;
			fh = FindFirstFile(dirpath, &FindFileData ); // the file returned is normally "." (not on root dir)

			if(fh == INVALID_HANDLE_VALUE )
			{
				printf("Invalid File Handle. GetLastError reports %d\n", GetLastError() );
				return 1; // unexpected error
			}
		}
		else
		{
			success = FindNextFile(fh, &FindFileData);

			if( ! success )
			{
				
				if( GetLastError() == ERROR_NO_MORE_FILES ) break;
				else printf("Error. GetLastError reports %d\n", GetLastError() );
				return 2; // unexpected error
			}
		}

		// delete if cFileName is != "." && != ".."
		if( strcmp(FindFileData.cFileName, ".") && strcmp(FindFileData.cFileName, "..") )
		{
			strcpy(delpath, path); // reset delpath to dir path, then attach filename
			PathMerger(delpath, FindFileData.cFileName);

			/* //DEBUG code
			printf("The first file found is %s\n", FindFileData.cFileName);
			printf("WARNING! Press any key to delete %s\n\n", delpath);
			getch();
			*/
			
			success = DeleteFile(delpath);

			//if( ! success ) return 3; // could not remove file. may be protected or in use

			 //DEBUG code
			if( ! success )
			{
				printf("Error. GetLastError reports %d\n", GetLastError() );
				FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL );
				printf("%s\n", lpMsgBuf );
							
			}
			//printf("path: %s\ndirpath: %s\ndelpath: %s\n", path, dirpath, delpath);
			//printf("FindFileData.cFileName: %s\n\n", FindFileData.cFileName);
			
		}
	}
	
	if( FindClose(fh) == 0 ) return 4; // Close file search. / Unexpected Error

	free(delpath);
	free(dirpath);

	return 0;
}



// deletes a file
// returns 0 on success, 1 on error
int DelFile(char *path)
{
	int success;
	
	// debug
	LPVOID lpMsgBuf;

	success = DeleteFile(path);
	if( ! success ) // could not remove file. may be protected or in use
	{
		printf("Error. GetLastError reports %d\n", GetLastError() );
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL );
		printf("%s\n", lpMsgBuf );
		
		return 1;
	}

	return 0;
}



// merges to parts of a path. Adds a directory separator if needed. Cuts of double DIR_SEPs (like in "c:/" + "/file" )
// path is ment to be long enough to hold the resulting path.
void PathMerger(char *path, char *filename)
{
	if( path[strlen(path)-1] != DIR_SEP_C ) strcat(path, DIR_SEP);

	if( filename[0] == DIR_SEP_C )			strcat(path, filename+1);
	else									strcat(path, filename);

	return;
}



// converts a string to lower case
int GetLowercase(char **lower, char *source)
{
	unsigned int i;
	
	*lower = realloc(*lower, strlen(source)+2);
	if ( *lower == NULL ) return 1;

	strcpy(*lower, source);

	for (i=0; i < strlen(*lower); i++)
	{
		if (((*lower)[i] >= 'A') && ((*lower)[i] <= 'Z')) (*lower)[i] += 32;
	}

	return 0;
}
