// PatchWorks 1.5 Image GetAverageColor File
// This Software is released under the terms of the GPL.

// calculates the average color of a bitmap



#include "_includes.h"
#include "_defines.h"
#include "_structs.h"




int GetAverageColor(sImage *image)
{
	uint i;
	uint imgsize;

	double avgc_b = 0;
	double avgc_g = 0;
	double avgc_r = 0;

	//imgsize = sizeof(image->data);
	//if(imgsize%3 != 0)					return 1;	// illegal image data size
	//if(imgsize != 3*image->w*image->h)	return 1;
	imgsize = 3*image->w*image->h;

	// this implementation works for images with
	// up to 16843009 pixels (4104*4104 pixels) -> PATCH_RES_MAXIMUM

	for(i=0; i<imgsize; i++)
	{
		switch(i%3)
		{
			case 0:	avgc_b += image->data[i]; break;
			case 1:	avgc_g += image->data[i]; break;
			case 2:	avgc_r += image->data[i]; break;
		}
	}

	imgsize /= 3;		// size of one channel

	image->avgc_b = (int)(avgc_b / imgsize + 0.5); // add 0.5 to floor/ceil
	image->avgc_g = (int)(avgc_g / imgsize + 0.5);
	image->avgc_r = (int)(avgc_r / imgsize + 0.5);

	return 0;
}




// Reads a Windows Bitmap (*.BMP) (Truecolor) into a sImage structure.
// Maintainer: Steven Schummer
//
// return values: 1 = file could not be read
//				  2 = not a bitmap file	
//                3 = not a 24bit bitmap file
//                4 = not enough memory
int ReadImage(sImage *srcImg, char *filename)
{
	FILE* fp;				// Filepointer
	ulong biSize=0;			// Header size - must be at least 40
	ulong biWidth=0;		// Image width in pixels
	ulong biHeight=0;		// Image height in pixels
	ushort biBM=0;			// Bitmap check
	ushort biPlanes=0;		// Must be 1
	ushort biBitCount=0;		// Bits per pixel - (1,2,4,8,16,24,32}
	ulong biCompression=0;		// Compression type (0 = uncompressed)
	ulong biSizeImage=0;		// Image size - uncompressed may be zero
	ulong biXPelsPerMeter=0;	// Preferred resolution (pixels/meter)
	ulong biYPelsPerMeter=0;	// Preferred resolution (pixels/meter)
	ulong sumByte=0;		// Size in bytes without headerinformation and fillbytes
	ulong fillBytes=0;		// Size in bytes of fillbytes in one row
	ulong biXdpi=0;			// Preferred resolution (pixels/inch)
	ulong biYdpi=0;			// Preferred resolution (pixels/inch)
	ulong i = 0;
	uint8 * readp;
	ulong count = 0;                // Counting variable for reading picture data
	ulong zCount = 0;		// Counting variable for reading picture data
	ulong pixRead = 0;		// Counting variable for reading picture data
	ulong bgrWidth = 0;		// Number of bytes in one row
	
	
	fp = fopen(filename, "rb"); // Opens a file (read binary)
	if(fp == NULL) return 1; // check if file could be read
	
	
	fseek(fp, 0, SEEK_SET);
	fread(&biBM, 2, 1, fp);
	
	if(biBM != 0x4D42) return 2;	//check for bitmap format, header entry
	
	fseek(fp, 28, SEEK_SET); // Go to the beginning of bitmapinfoheader
	fread(&biBitCount, 2, 1, fp); // Read bits per pixel value
	
	if(biBitCount != 24) return 3;
	
	fseek(fp, 14, SEEK_SET); // Go to beginning of info header
	fread(&biSize, 4, 1, fp); //Get headerinformation
	fread(&biWidth, 4, 1, fp);
	fread(&biHeight, 4, 1, fp);
	fread(&biPlanes, 2, 1, fp);
	fread(&biBitCount, 2, 1, fp);
	fread(&biCompression, 4, 1, fp);
	fread(&biSizeImage, 4, 1, fp);
	fread(&biXPelsPerMeter, 4, 1, fp);
	fread(&biYPelsPerMeter, 4, 1, fp);
	
	fseek(fp, 54, SEEK_SET); // Go to beginning of picture data
	
	// convert (pixel per meter) to (pixel per inch)
	biXdpi=(uint)(biXPelsPerMeter/(100/2.54f)+0.5); // +0.5 = rounding operation
	biYdpi=(uint)(biYPelsPerMeter/(100/2.54f)+0.5);
	
	// fill struct sImage with width, height, xdpi and ydpi
	srcImg->w=biWidth;
	srcImg->h=biHeight;
	srcImg->xdpi=biXdpi;
	srcImg->ydpi=biYdpi;
	
	sumByte = (srcImg->w * srcImg->h * 3); // Size in bytes without headerinformation and fillbytes
	fillBytes = (srcImg->w % 4); // Size in bytes of fillbytes in one row
	
	srcImg->data = malloc(sumByte); //allocate memory with sumByte bytes
	
	readp = srcImg->data;
	
	if(srcImg->data == NULL) return 4; //not enough memory
	
	bgrWidth = (srcImg->w) * 3; // number of bytes in one row
	
	// Read picture data in byte steps
	while(count < sumByte){
		while(zCount < bgrWidth) {
			pixRead = bgrWidth - zCount;
			zCount += fread(readp, 1, pixRead, fp);
			readp += zCount;
		}
		count += zCount;
		zCount = 0;
		fseek(fp, fillBytes, SEEK_CUR);
	}
	
	fclose(fp);
	
	return 0;
}




// Writes an Image from an sImage structure into a Windows Bitmap file (*.BMP)
// Maintainer: Steven Schummer
int WriteImage(sImage *srcImg, char *filename) {

	FILE* fp;
	ulong header4 = 0;	// 4 bytes for writing
	ushort header2 = 0;	// 2 bytes for writing
	uint8 leer = 0;		// Leerbyte
	ulong sumByte=0;	// Bytes without headerinformation and fillbytes
	ulong i = 0;
	ulong fillBytes=0;	// Bytes fillbytes
	uint8 * writep;		// picture data write pointer
	ulong count = 0;
	ulong zCount = 0;
	ulong bgrWidth = 0;
	ulong pixWrite = 0;
	ulong fileSize = 0;

	fillBytes = (srcImg->w % 4); // Get number of fillbytes per line
	sumByte = (srcImg->w * srcImg->h * 3); // Size in bytes of picture data
	bgrWidth = (srcImg->w) * 3; // Size of bytes of picture data per line

	writep = srcImg->data;
	
	fp = fopen(filename, "wb"); // Open file to write binary
	
	// Creating a bitmap fileformat
	// Description; Size 
	//
	// Signature; 2 bytes			 
	header2 = 0x4D42;
	fwrite(&header2,2,1,fp);

	// FileSize; 4 bytes
	fileSize = 54 + sumByte+fillBytes * srcImg->h;
	header4=fileSize+(fileSize%4);
	fwrite(&header4,4,1,fp);

	// Reserved; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// DataOffset; 4 bytes
	header4=54;
	fwrite(&header4,4,1,fp);

	// Size; 4 bytes
	header4=40;
	fwrite(&header4,4,1,fp);

	// Width; 4 bytes
	header4=srcImg->w;
	fwrite(&header4,4,1,fp);

	// Height; 4 bytes
	header4=srcImg->h;
	fwrite(&header4,4,1,fp);

	// Planes; 2 bytes
	header2=1;
	fwrite(&header2,2,1,fp);

	// BitCount; 2 bytes
	header2=24;
	fwrite(&header2,2,1,fp);

	// Compression; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// ImageSize; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	if( (srcImg->xdpi < PATCH_DPI_MINIMUM) || (srcImg->xdpi > PATCH_DPI_MAXIMUM) ) srcImg->xdpi = PATCH_DPI_DEFAULT;
	if( (srcImg->ydpi < PATCH_DPI_MINIMUM) || (srcImg->ydpi > PATCH_DPI_MAXIMUM) ) srcImg->ydpi = PATCH_DPI_DEFAULT;

	// XpixelsPerM; 4 bytes
	header4=(uint)((srcImg->xdpi*(100/2.54f))+0.5);
	fwrite(&header4,4,1,fp);

	// YpixelsPerM; 4 bytes
	header4=(uint)((srcImg->ydpi*(100/2.54f))+0.5);
	fwrite(&header4,4,1,fp);

	// ColorsUsed; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// ColorsImportant; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// picture data; 
	while(count < sumByte){
		while(zCount < bgrWidth) {
			pixWrite = bgrWidth - zCount;
			zCount += fwrite(writep, 1, pixWrite, fp);
			writep += zCount;
		}
		count += zCount;
		zCount = 0;
		for (i=0;i<fillBytes;i++) {
			fwrite(&leer, 1, 1, fp);
		}
	}
	// Add fillbytes at the end of picture data if necessary
	if(fileSize%4>0){
		for (i = 0; i < fileSize % 4; i++) {
			fwrite(&leer, 1, 1, fp);
		}
		
	}

	fclose(fp);
	return 0;
}
