/*
* PatchWorks 1.5 Image ScaleImage File
* This Software is released under the terms of the GPL.
* Scales an Image in a sImage structure
*/

/*
* ALGORITHMUS ZUR SKALIERUNG DER BILDER:
* Der Faktor mit dem die Werte der Farbkanle multipliziert werden, ndert
* sich je nach Breite und Hhe des bildes.
* Der Faktor von 3/5 entsteht aus "Breite des Zielbildes" / "Breite des Quellbilds"
* d.h. der Farbwert des ersten Pixels wird mit 3/5 multipliziert und mit 2/5
* des Farbwertes des zweiten Pixels addiert. Dies ergibt das 1. pixel des neuen Bildes.
* Jetzt muss der Farbwert des 2. pixels mit 1/5 multipliziert werden, das dritte
* Pixel mit 3/5 und das 4. Pixel mit 1/5. diese werte werden dann addiert. usw.
* Wird mit diesem Algorithmus gerechnet, so geht die Rechnung fr Bilder einer
* beliebigen Breite und Hhe auf.
* Beispiel:
* Im folgenden wird Quellbild mit qb abgekrzt und Zielbild mit zb
* Rx = Rotwert des x-ten Pixel
* Gx = Gelbwert des x-ten Pixel
* Bx = Blauwert des x-ten Pixel
* Breite des qb in Pixel: 13
* Breite des zb in Pixel: 5
* 1. pixel_zb = 5/13 * R1_qb + 5/13 * R2_qb + 3/13 * R3_qb
* 2. pixel_zb = 2/13 * R3_qb + 5/13 * R4_qb + 5/13 * R5_qb + 1/13 * R6_qb
* 3. pixel_zb = 4/13 * R6_qb + 5/13 * R7_qb + 4/13 * R8_qb
* 4. pixel_zb = 1/13 * R8_qb + 5/13 * R9_qb + 5/13 * R10_qb + 2/13 * R11_qb
* 5. pixel_zb = 3/13 * R11_qb + 5/13 * R12_qb + 5/13 * R13_qb
* Breite des qb in Pixel: 5
* Breite des zb in Pixel: 3
* 1. pixel_zb = 3/5 * R1_qb + 2/5 * R2_qb
* 2. pixel_zb = 1/5 * R2_qb + 3/5 * R3_qb + 1/5 * R4_qb
* 3. pixel_zb = 2/5 * R4_qb + 3/5 * R5_qb
* Breite des qb in Pixel: 8
* Breite des zb in Pixel: 4
* 1. pixel_zb = 4/8 * R1_qb + 4/8 * R2_qb
* 2. pixel_zb = 4/8 * R3_qb + 4/8 * R4_qb
* 3. pixel_zb = 4/8 * R5_qb + 4/8 * R6_qb
* 4. pixel_zb = 4/8 * R7_qb + 4/8 * R8_qb
* Bedingungen:
* - Zhler des Faktors darf nicht grer sein als dst->h
* - die Faktoren zur Berechnung eines Pixels des zbs mssen addiert 1 ergeben.
*
* Die Rechnung geht fr beliebeige Zahlenwerte immer auf.
* Der Algorithmus kann leicht abgendert auch zur Skalierung in der Hhe der
* Bilder verwendet werden.
*/


#include "_includes.h"
#include "_defines.h"
#include "_structs.h"




// skaliert Bild in Breite
int scaleWidth(sImage *src, sImage *dst);

// skaliert Bild in Hhe
int scaleHeight(sImage *src, sImage *dst);




int ScaleImage(sImage* srcImg, sImage* dstImg) {
	// dst->w / h enthalten bereits die Auflsung des neuen Bildes.
	sImage tmpImg; // wird zur Zwischenspeicherung, des in der Breite skalierten Bildes bentigt
	sImage tmpHack;// needed that 1:1 scales  work

	tmpImg.w=dstImg->w;
	tmpImg.h=srcImg->h;

	tmpImg.data=malloc(tmpImg.w*tmpImg.h*3);
	if( tmpImg.data == NULL ) return 1; // not enough memory

	// hack
	tmpHack = tmpImg;

	dstImg->data=malloc(dstImg->w*dstImg->h*3);
	if( dstImg->data == NULL ) return 1; // not enough memory

	if( srcImg->w < dstImg->w ) return 2; // src image too small. (can only downscale)
	if( srcImg->h < dstImg->h ) return 2;

	// scale width
/*	if( srcImg->w == dstImg->w ) tmpImg.data = srcImg->data;
	else	*/					 scaleWidth(srcImg, &tmpImg);

	// scale height
/*	if( srcImg->h == dstImg->h ) dstImg->data = tmpImg.data;
	else	*/					 scaleHeight(&tmpImg, dstImg);

	//free(tmpImg.data);
	//hack
	free(tmpHack.data);

	return 0;
}




int scaleWidth(sImage *src, sImage *dst) {
	uint i=0;
	uint counter=0;
	uint lastCounter=0;
	uint totalCounter=0;
	uint dstPixelPosition=0;
	uint srcPixelPosition=0;
	uint countSrcPixelFlag=0; // 0: dont count, or 1: count
	double dstPixelData[3]={0,0,0};
	double multiplier = 0;

	while(dst->w*dst->h != i) {
		if(totalCounter+dst->w < src->w && lastCounter==0) {
			counter=dst->w;
			lastCounter=0;
			countSrcPixelFlag=1;
		} else {
			if(lastCounter==0) {
				counter=src->w-totalCounter;
				lastCounter=counter;
				countSrcPixelFlag=1;
			} else {
				counter=dst->w-lastCounter;
				lastCounter=0;
			}
		}
		totalCounter+=counter;
		multiplier=(double)counter/src->w;
		dstPixelData[0]+=(double)multiplier*src->data[srcPixelPosition];
		dstPixelData[1]+=(double)multiplier*src->data[srcPixelPosition+1];
		dstPixelData[2]+=(double)multiplier*src->data[srcPixelPosition+2];

		if(countSrcPixelFlag==1) {
			srcPixelPosition+=3;
			countSrcPixelFlag=0;
		}

		if(totalCounter==src->w)
		{
			dst->data[dstPixelPosition]=(uint8)dstPixelData[0];
			dst->data[dstPixelPosition+1]=(uint8)dstPixelData[1];
			dst->data[dstPixelPosition+2]=(uint8)dstPixelData[2];
			dstPixelData[0]=0;
			dstPixelData[1]=0;
			dstPixelData[2]=0;
			totalCounter=0;
			counter=0;
			i++;
			dstPixelPosition+=3;
		}
		if(i%dst->w==0 && i != 0 && totalCounter==0) {
			lastCounter=0;
		}
	}
	return 0;
}




int scaleHeight(sImage *src, sImage *dst) {
	uint i=0;
	uint counter=0;
	uint lastCounter=0;
	uint totalCounter=0;
	uint dstPixelPosition=0;
	uint pixelPositionFirstRow=0;
	uint srcPixelPosition=0;
	uint countSrcPixelFlag=0; // 0: dont count, or 1: count
	double dstPixelData[3]={0,0,0};
	double multiplier = 0;

	while(dst->h*dst->w!=i) {
		if(totalCounter+dst->h < src->h && lastCounter==0) {
			counter=dst->h;
			lastCounter=0;
			countSrcPixelFlag=1;
		} else {
			if(lastCounter==0) {
				counter=src->h-totalCounter;
				lastCounter=counter;
				countSrcPixelFlag=1;
			} else {
				counter=dst->h-lastCounter;
				lastCounter=0;
			}
		}
		totalCounter+=counter;
		multiplier=(double)counter/src->h;
		dstPixelData[0]+=(double)multiplier*src->data[srcPixelPosition];
		dstPixelData[1]+=(double)multiplier*src->data[srcPixelPosition+1];
		dstPixelData[2]+=(double)multiplier*src->data[srcPixelPosition+2];

		if(countSrcPixelFlag==1) {
			srcPixelPosition+=3*src->w;
			countSrcPixelFlag=0;
		}
		if(totalCounter==src->h)
		{
			dst->data[dstPixelPosition]=(uint8)dstPixelData[0];
			dst->data[dstPixelPosition+1]=(uint8)dstPixelData[1];
			dst->data[dstPixelPosition+2]=(uint8)dstPixelData[2];
			dstPixelData[0]=0;
			dstPixelData[1]=0;
			dstPixelData[2]=0;
			counter=0;
			totalCounter=0;
			i++;
			dstPixelPosition+=3*dst->w;
		}
		if(i%dst->h==0 && i != 0 && totalCounter==0) {
			pixelPositionFirstRow+=3;
			dstPixelPosition=pixelPositionFirstRow;
			srcPixelPosition=pixelPositionFirstRow;
			lastCounter=0;
		}
	}
	return 0;
}
