// PatchWorks 1.5 Patchwork CreatePatchwork File
// This Software is released under the terms of the GPL.

// Assembles a Patchwork, using the currently open Databases and the source image.



#include "Patchwork.h"

int FindMatchingPatch(uint, uint, uint, sDataBase [], char **, uint *, uint tolerance);

int CreatePatchwork(char *srcimg, char *dstimg, sDataBase DataBase[], uint print, double tolerancepercent)
{
	sImage srcImg;
	sImage dstImg;
	sImage patImg; // temporary patch structure

	uint tolerance = (uint)((ASSESSMENT_MAX * tolerancepercent) / 100); // tolerancepercent= 0.0...100, tolerance= 0...3*255
	uint missmatch;	// how proper a patch matches the replaced pixel
	uint minmissmatch = ASSESSMENT_MAX;
	uint maxmissmatch = 0;
	double avgmissmatch = 0;

	char *patchfile;
	
	uint r, g, b; // rgb-Values of source pixels
	uint i, j, k, l;
	uint err, dbinuse;
	uint srcoff, dstoff, dstoff_w, dstoff_h, dstoff_t, patoff;
	uint patch_w, patch_h;
	double percentage;
	uint srcpixels, calcedpixels;
	uint used, allused = 0, allimagecount = 0;



	// read source image
	err = ReadImage(&srcImg, srcimg);
	if( err ) return 10 + err; // error while reading the source image

	// find a DB thats in use
	dbinuse = -1;
	for(i=0; i<MAX_OPEN_DBs; i++)
	{
		if(DataBase[i].inuse == 1)
		{
			dbinuse=i;
			break;
		}
	}
	if( dbinuse == -1 ) return 2;		// No DB in use


	// Patch resolution
	patch_w = DataBase[dbinuse].res_w;
	patch_h = DataBase[dbinuse].res_h;

	// Size of destination picture
	dstImg.w=srcImg.w*patch_w;
    dstImg.h=srcImg.h*patch_h;

	dstImg.data=malloc(dstImg.w*dstImg.h*3);	
	if(dstImg.data == NULL)	return 3;	// not enough memory


	// percentage
	if(print) printf("\t  0.00%%");
	srcpixels = srcImg.w * srcImg.h;
	calcedpixels = 0;

	// do conversion
	for(i=0; i<srcImg.h; i++) // vertical patch count
	{
		for(j=0; j<srcImg.w; j++) // hozizontal patch count
		{
			srcoff = (srcImg.w*i + j) * 3; //full lines plus current offset times 3 bytes

			b = srcImg.data[srcoff  ];
			g = srcImg.data[srcoff+1];
			r = srcImg.data[srcoff+2];

			err = FindMatchingPatch(r, g, b, DataBase, &patchfile, &missmatch, tolerance);
			if( err ) return 20 + err;

			// missmatch statistics
			avgmissmatch += missmatch; // would work for at least 5.614.336 sourcepixels if int32, more for double
			if( missmatch < minmissmatch) minmissmatch = missmatch;
			if( missmatch > maxmissmatch) maxmissmatch = missmatch;


			// percentage
			calcedpixels++;
			percentage = 100 * calcedpixels / (double)srcpixels;
			if( !(calcedpixels%200) && print ) printf("\b\b\b\b\b\b\b%6.2f%%", percentage);

			err = ReadImage(&patImg, patchfile);
			if( err ) return 30 + err; // error while reading patch image
			free(patchfile);

			dstoff_h = i*srcImg.w*patch_w*patch_h*3;// total offset of fully-written patchlines
			dstoff_w = patch_w * j * 3;
			dstoff_t = dstoff_h + dstoff_w; // offset of FIRST byte of current patch

			for(k=0; k<patch_h; k++)
			{
				dstoff = dstoff_t + k*dstImg.w*3; // add linewith * writtenlines
				patoff   = k*patch_w*3;

				for(l=0; l<patch_w*3; l++)
				{
					dstImg.data[dstoff+l] = patImg.data[patoff + l];
				}
			}

			free(patImg.data);
		}
	}

	if(print)
	{
		printf("\b\b\b\b\b\b\b100.00%\n");
		printf(eWRITING_IMAGE, dstimg);
	}

	// write destination image
	err = WriteImage(&dstImg, dstimg);
	if( err ) return 40 + err; // error while reading the source image

	free(srcImg.data);
	free(dstImg.data);

	// do some statistics
	if(print)
	{
		printf(eSTATISTICS "\n");

		avgmissmatch /= (srcImg.w * srcImg.h); // divide with source image pixel count

		printf(eSTATISTIC_AVGMM, avgmissmatch, (100.0 * avgmissmatch) / ASSESSMENT_MAX);
		printf(eSTATISTIC_MINMM, minmissmatch, (100.0 * minmissmatch) / ASSESSMENT_MAX);
		printf(eSTATISTIC_MAXMM, maxmissmatch, (100.0 * maxmissmatch) / ASSESSMENT_MAX);

		printf("\n");

		for(k=0; k<MAX_OPEN_DBs; k++)
		{
			if(DataBase[k].inuse == 0) continue;  // try next DB
			
			used = 0;
			for(i=0; i<DataBase[k].imgcount; i++)
			{
				if( DataBase[k].usecount[i] ) used++;
			}
			allused       += used;
			allimagecount += DataBase[k].imgcount;
			
			printf(eSTATISTIC_ONE, used, DataBase[k].imgcount, DataBase[k].path);
		}
		printf("\n     ==============   ");
		printf(eSTATISTIC_ALL, allused, allimagecount);
	}
	return 0;
}



typedef struct
{
	uint dbnumber;
	uint imgnumber;
	//double assessment;
	uint assessment;
	uint usecount;
}sMatching;


// returns the path of the best matching patch to a given pattern.
#if defined(_WIN32) || defined(__WIN32__)
__inline int FindMatchingPatch(uint r, uint g, uint b, sDataBase DataBase[], char **matchingpath, uint *missmatch, uint tolerance)
#else
		 int FindMatchingPatch(uint r, uint g, uint b, sDataBase DataBase[], char **matchingpath, uint *missmatch, uint tolerance)
#endif
{
	uint k, i, o;
	// uint err;
	//uint rtmp, gtmp, btmp; // temp colors rgb
	uint *col; // shortcut for DataBase[k].colordata
	uint matching, bestmatching, dbnum; // bestmatching: matches best, but matching may be used less often
	uint toleratedassessment;			// how much an alternative patch can be away from the one with the best assessment

	//double *assessment = NULL; // assessment, how propper a color fits the given pattern. (int would be faster)
	uint *assessment = NULL; // NULL initialized for realloc

	char filename[INDEX_FILE_LINE]; // long enough...
	int filenameoff;
	
	sMatching candidate[MAX_OPEN_DBs];
	uint candidatecount=0;	// counts candidate (= count of open DBs)

	for(k=0; k<MAX_OPEN_DBs; k++)
	{
		if(DataBase[k].inuse == 0) continue;  // try next DB

		assessment = realloc(assessment, INDEX_FILE_LINE * DataBase[k].imgcount);
		if( assessment == NULL ) return 1; // not enough memory

		col = DataBase[k].colordata; // shortcut

		bestmatching = 0;	// set to first element at the beginning
		
		for(i=0; i<DataBase[k].imgcount; i++)
		{
			/*	// (1a.) slowest (with sscanf)
			o = i*INDEX_FILE_LINE + 13;		// offset. 13 = first byte of RRR, GGG, BBB
			err = sscanf(&DataBase[k].data[o], "%i %i %i", &rtmp, &gtmp, &btmp);
			if(err != 3) return 2;				// could not read all channels
				// (1b) double as fast (with atoi)
			rtmp = atoi(&DataBase[k].data[o]);
			gtmp = atoi(&DataBase[k].data[o+4]);
			btmp = atoi(&DataBase[k].data[o+8]);

			if( rtmp > r ) rtmp = rtmp - r;		// difference in r channel
			else		   rtmp = r - rtmp;
			if( gtmp > g ) gtmp = gtmp - g;		// difference in g channel
			else		   gtmp = g - gtmp;
			if( btmp > b ) btmp = btmp - b;		// difference in b channel
			else		   btmp = b - btmp;
			assessment[i] = rtmp + gtmp + btmp;
			*/
			/*	// (2) little bit faster (number conversion done in OpenDB)
			o = i*3;
			rtmp = col[o  ];
			gtmp = col[o+1];
			btmp = col[o+2];
			if( rtmp > r ) rtmp = rtmp - r;		// difference in r channel
			else		   rtmp = r - rtmp;
			if( gtmp > g ) gtmp = gtmp - g;		// difference in g channel
			else		   gtmp = g - gtmp;
			if( btmp > b ) btmp = btmp - b;		// difference in b channel
			else		   btmp = b - btmp;
			assessment[i] = rtmp + gtmp + btmp;
			// very slow, no quality benefits (and ment for assessment to be double)
			//assessment[i] = pow(2, rtmp) + pow(2, gtmp) + pow(2, btmp);
			*/

			// (3) fastest (number conversion done in OpenDB, less additions)
			o = i*3;

			if( col[o  ] > r )	assessment[i]  = col[o  ] - r;		// difference in r channel
			else				assessment[i]  = r - col[o  ];
			if( col[o+1] > g )	assessment[i] += col[o+1] - g;		// difference in g channel
			else				assessment[i] += g - col[o+1];
			if( col[o+2] > b )	assessment[i] += col[o+2] - b;		// difference in b channel
			else				assessment[i] += b - col[o+2];

			// check if patch matches better than previous patches.
			if(	 assessment[i] <     assessment[bestmatching] ) bestmatching = i;
		}

		matching			= bestmatching;
		toleratedassessment = tolerance + assessment[bestmatching];

		// if bestmatching is used too often, take a slightly worse matching
		for(i=0; i<DataBase[k].imgcount; i++)
		{
			// <= is used (insteaed of < ) so a patch with same assessment and a less usecount is preferred
			if(	(assessment[i] <= toleratedassessment) && (DataBase[k].usecount[i] < DataBase[k].usecount[bestmatching]) ) matching = i;
		}

		// add to candidates
		candidate[candidatecount].dbnumber   = k;
		candidate[candidatecount].imgnumber  = matching;
		candidate[candidatecount].assessment = assessment[matching];
		candidate[candidatecount].usecount   = DataBase[k].usecount[matching];
		candidatecount++;

	} // for all DBs


	// finally choose REALLY best matching (out of the best matches of all used DBs)
	matching = 0; // set to first element at the beginning
	bestmatching = 0;

	for(i=1; i<candidatecount; i++)
	{
		//if( candidate[i].assessment < candidate[bestmatching].assessment ) bestmatching = i;
		if( (candidate[i].assessment <= candidate[matching].assessment) && (candidate[i].usecount < 5 + candidate[matching].usecount) ) matching = i;

	}

	//get missmatch to return
	*missmatch = candidate[matching].assessment;

	// get path to return
	dbnum		= candidate[matching].dbnumber;
	filenameoff	= candidate[matching].imgnumber * INDEX_FILE_LINE;

	*matchingpath = malloc( strlen(DataBase[dbnum].path) + INDEX_FILE_LINE );
	if( *matchingpath == NULL ) return 1; // not enough memory

	for(k=0; k<12; k++)	filename[k] = DataBase[dbnum].data[filenameoff+k];
	filename[12]='\0';

	strcpy(*matchingpath, DataBase[dbnum].path);

	PathMerger(*matchingpath, filename);

	// increase number of usings
	DataBase[dbnum].usecount[ candidate[matching].imgnumber ]++;

	free(assessment);
//	free(matchingpath); // NOT to be freed...

	return 0;
}