// PatchWorks 1.5 Main Console File
// This Software is released under the terms of the GPL.

// Console Interface


#include "_includes.h"
#include "_defines.h"
#include "_structs.h"

#include "AdminDB.h"
#include "Image.h"
#include "Patchwork.h"
#include "FileSysHelp.h"


void Help()
{
	printf(help);
}


int Console(int argc, char *argv[])
{
	sDataBase DataBase[MAX_OPEN_DBs] = {0};		// a maximum of MAX_OPEN_DBs Databases can be open at the same time
	int i, err, resx, resy, dpi, cC;
	char answer;
	char *argaction = NULL, *argmode = NULL;
	clock_t starttime, stoptime, time_ms;
	double time_s;
	double tolerancepercent;

		// get lowercase parameter for stringcompare
					err  = GetLowercase(&argaction, argv[1]);	// action parameter (-addDB, -subDB, -newDB, -delDB)
	if(argc > 3)	err += GetLowercase(&argmode,   argv[3]);	// mode   parameter (force / pseudo)
	if( err )
	{
		printf(eNO_MEMORY);
		return err;
	}


	if (strcmp(argaction , "-deldb" ) ==0)
	{
		if(argc<3)								// to few parameters for deldb
		{
			printf(err2);	return 1;
		}

		if(argc==3)
		{
			do
			{
				fflush(stdin);
				printf(err3);					// delete DB? Y/N
				scanf("%c",&answer);
			}
			while ((answer!='y') && (answer!='Y') && (answer!='n') && (answer!='N'));

			if ((answer=='n')||(answer=='N'))
			{
				printf(err9_DeleteDB);			// user abort
				return 1; 
			}
		}
		if(argc==4)
		{
			// get lowercase parameter for stringcompare
			err = GetLowercase(&argaction, argv[3]);
			if( err )
			{
				printf(eNO_MEMORY);
				return err;
			}
			if ( strcmp(argaction,"force") )		// wrong parameter (force)
			{
				printf(err5,argc);
				return 1;
			}
		}
		if(argc > 4)							// too many parameters
		{
			printf(err28);	return 1;
		}

		// open DB
		err=OpenDB(DataBase, argv[2]);			// openDB error 1-3 or 11-18
		if ((err>10)&&(err<20))					// ReadIndexFile errors 11-18
		{
			if(err == 19) printf(err32, DB_INDEX_FILE, DB_INDEX_VERSION);		// wrong DataBase Version
			printf(err6,err);					//readindex failure
			return 1;
		}
		else
		{
			switch( err )
			{
			case 1:printf(err1_OpenDB);return 1;// To many DBs open
			case 2:printf(eNO_MEMORY);return 1;	// not enough memory?
			case 3:printf("OpenDB: " eNOT_A_DIR);return 1;	// could not access directory
			default: break;						//db opened
			}
		}

		// delete DB
		err=DeleteDB(DataBase, 0);				// DeleteDB errors 1-4 or 11-12
		if ((err>10)&&(err<20))					// closeDB errors
		{
			printf(err3_CloseDB, err);			//closeDB failure code %i
			return 1;
		}
		else
		{
			switch( err )
			{
			case 0:printf(err0_DeleteDB);break;		// success
			case 1:printf(err1_DeleteDB);return 1;	// already closed / not in use
			case 2:printf(err2_DeleteDB);return 1;	// unsaved changes! save, or set changed=0 to override
			case 3:printf(err3_DeleteDB);return 1;	// could not delete (all) items in folder
			case 4:printf(err4_DeleteDB);return 1;	// could not remove folder
			default:printf(eINTERNAL_ERROR); return 1;
			}
		}
	}

	//########################################################################################################

	else if (strcmp ( argaction , "-newdb" ) ==0)
	{
		resx = PATCH_RES_DEFAULT;
		resy = PATCH_RES_DEFAULT;
		dpi  = PATCH_DPI_DEFAULT;

		if(argc <  3)									// too few parameters
		{
			printf(err1);	return 1;
		}
										// if argc == 3, default values are used
		if(argc >= 4)									// x==y, dpi=default
		{
			err = sscanf(argv[3],"%i",&resx);
			if( err != 1 )
			{
				printf(err8, argv[3]);	return 1;
			}
			resy = resx;
		}
		if(argc>=5)										// x, y, dpi=default
		{
			err = sscanf(argv[4],"%i",&resy);
			if( err != 1 )
			{
				printf(err8, argv[4]);	return 1;
			}
		}
		if(argc==6)										// x, y, dpi
		{
			err = sscanf(argv[5],"%i",&dpi);
			if( err != 1 )
			{
				printf(err8, argv[5]);	return 1;
			}
		}
		if(argc > 6)									// too many parameters
		{
			printf(err28);	return 1;
		}

		// add DataBase
		err=AddDB(argv[2], resx, resy, dpi);
		if ((err>10)&&(err<20))						// createIndexFile errors 11-12
		{
			printf(err7,err);						//createIndexFile failure
			return 1;
		}
		else
		{
			switch( err )
			{
			case 0:printf(err0_AddDB);break;		// success
			case 1:printf(err1_AddDB);return 1;		// illegal resolution
			case 2:printf(err2_AddDB);return 1;		// not enough meomory
			case 3:printf(err3_AddDB);return 1;		// already exists
			case 4:printf(err4_AddDB);return 1;		// couldn't create
			default:printf(eINTERNAL_ERROR); return 1;
			}
		}
	}

	//#####################################################################################################
	else if (strcmp ( argaction , "-subdb" ) ==0)
	{
		if (argc<4)									// not enough parameters for -subdb
		{
			printf(err9);	return 1;
		}
		if(argc==4)									// normal call
		{
			do
			{
				printf(err21);						// ask user to confirm delete DB item
				scanf("%c",&answer);
			}
			while ((answer!='y') && (answer!='Y') && (answer!='n') && (answer!='N'));

			if ((answer=='n')||(answer=='N'))		// return without delete DB item
			{
				printf(err22);						// DB item not deleted
				return 1; 
			}
		}
		if(argc==5)									// call with "force"
		{
			if (strcmp (argmode,"force") != 0)		// check parameter "force"
			{
				printf(err5,argc);
				return 1;
			}
		}
		if(argc > 5)								// too many parameters
		{
			printf(err28); return 1;
		}

		//open and delete DB item
		err=OpenDB(DataBase, argv[2]);				// openDB error 1-3 or 11-18
		if ((err>10)&&(err<20))						// readindexfile errors 11-18
		{
			if(err == 19) printf(err32, DB_INDEX_FILE, DB_INDEX_VERSION);		// wrong DataBase Version
			printf(err6,err);	// readindex failure
			return 1;
		}
		else
		{
			switch( err )
			{
			case 1:printf(err1_OpenDB);return 1;	// Too many DBs open
			case 2:printf(eNO_MEMORY);return 1;		// not enough memory?
			case 3:printf("OpenDB: " eNOT_A_DIR);return 1;		// could not access directory
			default: break;							// db opened
			}
		}

		err=DeleteDBItem(DataBase, argv[3]);
		if ((err>10)&&(err<20))						// closeDB errors 11-18
		{
			printf(err3_CloseDB,err);				// closeDB failure
			return 1;
		}
		else
		{
			switch( err )
			{
			case 1:printf(err1_DeleteDBItem);return 1;//

			default:printf(err10);
			}
		}
	}

	//###########################################################################################################

	else if (strcmp ( argaction , "-adddb" ) ==0)
	{
		if (argc < 4)									// not enough parameters for adddb
		{
			printf(err11);	return 1;
		}
		if (argc > 4)									// too many parameters
		{
			printf(err28); return 1;
		}
		
		// open DB
		err=OpenDB(DataBase, argv[2]);					// openDB error 1-3 or 11-18
		if ((err>10)&&(err<20))							// readindexfile errors 11-18
		{
			if(err == 19) printf(err32, DB_INDEX_FILE, DB_INDEX_VERSION);		// wrong DataBase Version
			printf(err6,err);	// readindex failure
			return 1;
		}
		else
		{
			switch( err )
			{
			case 1:printf(err1_OpenDB); return 1;		// Too many DBs open
			case 2:printf(eNO_MEMORY);  return 1;		// not enough memory?
			case 3:printf("OpenDB: " eNOT_A_DIR);  return 1;		// not a directory
			default: break;								// db opened
			}
		}

		// add item(s) to DB
		err=AddDBItems(argv[3], DataBase, 1);
		if((err>10)&&(err<20))				// ReadImage errors
		{
			printf(err12,err); return 1;
		}
		if((err>20)&&(err<30))				// ScaleImage errors
		{
			printf(err13,err); return 1;
		}
		if((err>30)&&(err<40))				// GetAverageColor errors
		{
			printf(err14,err); return 1;
		}
		if((err>40)&&(err<50))				// WriteImage errors
		{
			printf(err15,err); return 1;
		}

		switch( err )
		{
		case 1: printf(err1_AddDBItem);return 1;	// not enough memory
		case 2: printf(err2_AddDBItem);return 1;	// illegal database specified
		case 3: printf(err3_AddDBItem);return 1;	// DB is full
		case 4: printf(err4_AddDBItem);return 1;	// internal error
		case 8: printf(err1_AddDBItem);return 1;	// not enough memory
		case 9: printf(err9_AddDBItem);
		default:printf(err0_AddDBItem);
		}

		err=SaveDB(DataBase, 0);
		if ((err>10)&&(err<20))					// CreateIndexFile errors
		{
			printf(err7,err);	return 1;
		}

		switch (err)
		{
		case 1:printf(err1_SaveDB);return 1;	// DB not in use
		case 2:printf(err2_SaveDB);return 1;	// no changes to save
		case 3:printf(err3_SaveDB);return 1;	// not enough memory?
		case 4:printf(err4_SaveDB);return 1;	// directory does not exist?
		case 5:printf(err5_SaveDB);return 1;	// could not delete file
		case 6:printf(err6_SaveDB);return 1;	// could not create file
		default:printf(err0_SaveDB);
		}
	}

	//##############################################################################################################

	else // (PSEUDO) Patchwork
	{
		if(argc <  4)									// too few parameters
		{
			printf(err1);	return 1;
		}
		if(argc >  4 + MAX_OPEN_DBs)					// too many parameters
		{
			printf(err18);	return 1;
		}
		
		if( strcmp ( argmode, "pseudo" ) ==0 )
		{
			if(argc >  7)								// too many parameters for pseudo
			{
				printf(err16); return 1;
			}
			// we have 4, 5 or 6 parameters

			cC   = COLORCOUNT_DEFAULT;
			resx = PATCH_RES_DEFAULT;
			resy = PATCH_RES_DEFAULT;

			if(argc > 4)	// colorcount is given
			{
				err = sscanf(argv[4],"%i",&cC);
				if( err != 1 )
				{
					printf(err8, argv[4]);	return 1;
				}
			}
			if(argc > 5)	// resolution is given (for x or x & y)
			{
				sscanf(argv[5],"%i",&resx);
				if( err != 1 )
				{
					printf(err8, argv[5]);	return 1;
				}
				if ( (resx < PATCH_RES_MINIMUM) || ( resx > PATCH_RES_MAXIMUM) )
				{
					printf(eILLEGAL_RESOLUTION);
					return 1;
				}

				resy = resx;
			}
			if(argc > 6)	// resolution is given (for y)
			{
				sscanf(argv[6],"%i",&resy);
				if( err != 1 )
				{
					printf(err8, argv[6]);	return 1;
				}
				if ( (resy < PATCH_RES_MINIMUM) || ( resy > PATCH_RES_MAXIMUM) )
				{
					printf(eILLEGAL_RESOLUTION);
					return 1;
				}
			}

			printf(eSTART_PSEUDO, cC, resx, resy);			// start calculation
			starttime = clock();
			
			err=CreatePseudoPatchwork(argv[1], argv[2], resx, resy, cC);
			switch (err)
			{
			case 1:printf(eNO_MEMORY);return 1;	// not enough memory
			default:printf(err17);break;
			}
			
			stoptime = clock();
			time_ms = stoptime - starttime;
			time_s = time_ms / 1000.0;
			printf(eTIME_ELAPSED, time_s);
		}
		//#########################################################################################################

		else // not pseudo, normal patchwork
		{
			// open DBs (max. 10)
			for (i=4; i<argc; i++)
			{
				err=OpenDB(DataBase, argv[i]);	// openDB error 1-3 or 11-18
				if ((err>10)&&(err<20))			// readindexfile error 11-18
				{
					if(err == 19) printf(err32, DB_INDEX_FILE, DB_INDEX_VERSION);		// wrong DataBase Version
					printf(err6,err);			// readindex failure
					return 1;
				}
				else 
				{
					switch( err )
					{
					case 1:printf(err1_OpenDB);return 1;	// Too many DBs open
					case 2:printf(eNO_MEMORY);return 1;		// not enough memory?
					case 3:printf("OpenDB: " eNOT_A_DIR);return 1;		// not a directory
					default: printf(err20);break;			// db opened
					}
				}				
			}

			for (i=1; i<argc-4; i++)		// compare DB res_h and res_w, close if different
			{
				if( (DataBase[0].res_h != DataBase[i].res_h) || (DataBase[0].res_w != DataBase[i].res_w) )
				{
					printf(err29); return 1; // DBs don't fit
				}
			}

			err = sscanf(argv[3], "%lf", &tolerancepercent);
			if(err != 1)
			{
				printf(err30);
				return 1;
			}
			if( (tolerancepercent < 0.0) || (tolerancepercent > 100.0) )
			{
				tolerancepercent = TOLERANCE_DEFAULT;
				printf(err31, tolerancepercent);
			}

			printf(eSTART_PATCHWORK, argc-4, DataBase[0].res_w, DataBase[0].res_h); // start calculation
			starttime = clock();

			err=CreatePatchwork(argv[1], argv[2], DataBase, 1, tolerancepercent);
			if ((err>10)&&(err<20))				// ReadImage errors
			{
				printf(err12,err);	return 1;
			}
			if ((err>20)&&(err<30))				// FindFittingPatch errors
			{
				printf(err25,err);	return 1;
			}
			if ((err>30)&&(err<40))				// ReadImage errors (patchimages)
			{
				printf(err12,err);	return 1;
			}
			if ((err>40)&&(err<50))				// WriteImage errors (patchimages)
			{
				printf(err12,err);	return 1;
			}
			switch( err )
			{
			case 0: break;
			case 2:printf(err26);	return 1;	// No DB in use
			case 3:printf(eNO_MEMORY);return 1;	// not enough memory
			default:printf(eINTERNAL_ERROR);
			}

			stoptime = clock();
			time_ms = stoptime - starttime;
			time_s = time_ms / 1000.0;
			printf(eTIME_ELAPSED, time_s);
			printf(err27 "\n");
		} // if pseudo
	}

	free( argaction );
	free( argmode );
	return 0;
}